import os
import shutil

import audeer


# Project -----------------------------------------------------------------
project = 'audeer'
author = 'Hagen Wierstorf, Johannes Wagner'
version = audeer.git_repo_version()
title = '{} Documentation'.format(project)


# General -----------------------------------------------------------------
master_doc = 'index'
source_suffix = '.rst'
exclude_patterns = [
    'build',
    'tests',
    'Thumbs.db',
    '.DS_Store',
    'api-src',
]
templates_path = ['_templates']
pygments_style = None
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',  # support for Google-style docstrings
    'sphinx.ext.autosummary',
    'sphinx_autodoc_typehints',
    'sphinx_copybutton',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
]
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
}
linkcheck_ignore = [
    'https://gitlab.audeering.com',
]
copybutton_prompt_text = r'>>> |\.\.\. |$ '
copybutton_prompt_is_regexp = True

autodoc_default_options = {
    'undoc-members': False,
}

# Disable auto-generation of TOC entries in the API
# https://github.com/sphinx-doc/sphinx/issues/6316
toc_object_entries = False


# HTML --------------------------------------------------------------------
html_theme = 'sphinx_audeering_theme'
html_theme_options = {
    'display_version': True,
    'footer_links': False,
    'logo_only': False,
}
html_context = {
    'display_github': True,
}
html_title = title


# Copy API (sub-)module RST files to docs/api/ folder ---------------------
audeer.mkdir('api')
api_src_files = audeer.list_file_names('api-src')
api_dst_files = [
    audeer.path('api', os.path.basename(src_file))
    for src_file in api_src_files
]
for src_file, dst_file in zip(api_src_files, api_dst_files):
    shutil.copyfile(src_file, dst_file)



import os
import subprocess
import warnings

import pytest

import audeer


def test_deprecated():

    @audeer.deprecated(removal_version='1.0.0', alternative='audeer.mkdir')
    def deprecated_function():
        pass

    expected_message = (
        'deprecated_function is deprecated '
        'and will be removed with version 1.0.0.'
        ' Use audeer.mkdir instead.'
    )
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        deprecated_function()
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)

    @audeer.deprecated(removal_version='2.0.0')
    def deprecated_function():
        pass

    expected_message = (
        'deprecated_function is deprecated '
        'and will be removed with version 2.0.0.'
    )
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        deprecated_function()
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)

    @audeer.deprecated(removal_version='2.0.0')
    class DeprecatedClass():
        pass

    expected_message = (
        'DeprecatedClass is deprecated '
        'and will be removed with version 2.0.0.'
    )
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        DeprecatedClass()
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)


def test_deprecated_default_value():

    @audeer.deprecated_default_value(
        argument='foo',
        new_default_value='bar',
        change_in_version='1.0.0',
    )
    def function_with_deprecated_default_value(*, foo='foo'):
        return foo

    expected_message = (
        "The default of 'foo' will change from "
        "'foo' to 'bar' "
        "with version 1.0.0."
    )
    default_value = 'foo'
    with pytest.warns(None):
        # no warning if we set value
        function_with_deprecated_default_value(foo='foo')
        function_with_deprecated_default_value(foo='bar')
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        assert function_with_deprecated_default_value() == default_value
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)


def test_deprecated_keyword_argument():

    @audeer.deprecated_keyword_argument(
        deprecated_argument='foo',
        new_argument='bar',
        removal_version='1.0.0',
        mapping=lambda x: 2 * x,
    )
    def function_with_deprecated_keyword_argument(*, bar):
        return bar

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
        " Use 'bar' instead."
    )
    value = 1
    assert function_with_deprecated_keyword_argument(bar=value) == value
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=value)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 2 * value

    @audeer.deprecated_keyword_argument(
        deprecated_argument='foo',
        removal_version='1.0.0',
    )
    def function_with_deprecated_keyword_argument(**kwargs):
        return 1

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
    )
    assert function_with_deprecated_keyword_argument() == 1
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=2)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 1

    @audeer.deprecated_keyword_argument(
        deprecated_argument='foo',
        removal_version='1.0.0',
        remove_from_kwargs=False,
    )
    def function_with_deprecated_keyword_argument(**kwargs):
        if 'foo' in kwargs:
            return kwargs['foo']
        else:
            return 1

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
    )
    assert function_with_deprecated_keyword_argument() == 1
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=2)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 2

    @audeer.deprecated_keyword_argument(
        deprecated_argument='foo',
        new_argument='bar',
        removal_version='1.0.0',
        remove_from_kwargs=False,
    )
    def function_with_deprecated_keyword_argument(**kwargs):
        if 'foo' in kwargs:
            return kwargs['foo']
        else:
            return 1

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
        " Use 'bar' instead."
    )
    assert function_with_deprecated_keyword_argument() == 1
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = function_with_deprecated_keyword_argument(foo=2)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r == 2

    @audeer.deprecated_keyword_argument(
        deprecated_argument='foo',
        new_argument='bar',
        removal_version='1.0.0',
    )
    class class_with_deprecated_keyword_argument(object):

        def __init__(self, *, bar):
            self.bar = bar

    expected_message = (
        "'foo' argument is deprecated "
        "and will be removed with version 1.0.0."
        " Use 'bar' instead."
    )
    value = 1
    assert class_with_deprecated_keyword_argument(bar=value).bar == value
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Raise warning
        r = class_with_deprecated_keyword_argument(foo=value)
        assert issubclass(w[-1].category, UserWarning)
        assert expected_message == str(w[-1].message)
        assert r.bar == value


@pytest.mark.parametrize('nested_list,expected_list', [
    ([1, 2, 3, [4], [], [[[[[[[[[5]]]]]]]]]], [1, 2, 3, 4, 5]),
    ([[1, 2], 3], [1, 2, 3]),
    ([1, 2, 3], [1, 2, 3]),
])
def test_flatten_list(nested_list, expected_list):
    flattened_list = audeer.flatten_list(nested_list)
    assert flattened_list == expected_list


def test_freeze_requirements(tmpdir):
    path = str(tmpdir.mkdir('tmp'))
    outfile = os.path.join(path, 'requirements.txt.lock')
    audeer.freeze_requirements(outfile)
    with open(outfile) as f:
        requirements = f.readlines()
    # Remove whitespace and \n
    requirements = [r.strip() for r in requirements]
    assert any(['pytest' in r for r in requirements])
    assert any(['audeer' in r for r in requirements])
    with pytest.raises(RuntimeError, match=r'Freezing Python packages failed'):
        outfile = os.path.join(path, 'tmp2/requirements.txt.lock')
        audeer.freeze_requirements(outfile)


def test_git_repo_tags():
    git = ['git', 'tag']
    expected_tags = subprocess.check_output(git)
    expected_tags = expected_tags.decode().strip().split('\n')
    tags = audeer.git_repo_tags()
    assert tags == expected_tags
    tags = audeer.git_repo_tags(v=True)
    expected_tags = [
        f'v{t}' if not t.startswith('v') else t for t in expected_tags
    ]
    assert tags == expected_tags
    tags = audeer.git_repo_tags(v=False)
    expected_tags = [
        t[1:] if t.startswith('v') else t for t in expected_tags
    ]
    assert tags == expected_tags


def test_git_repo_version():
    git = ['git', 'describe', '--tags', '--always']
    expected_version = subprocess.check_output(git)
    expected_version = expected_version.decode().strip()
    version = audeer.git_repo_version(v=True)
    if not expected_version.startswith('v'):
        expected_version = f'v{expected_version}'
    assert version == expected_version
    version = audeer.git_repo_version(v=False)
    assert version == expected_version[1:]


@pytest.mark.parametrize(
    'version, is_semantic',
    [
        (
            '1.0.0',
            True,
        ),
        (
            'v1.0.0',
            True,
        ),
        (
            '4.0.0-20200206.095534-3',
            True,
        ),
        (
            'v1.0.1-1-gdf29c4a',
            True,
        ),
        (
            '1',
            False,
        ),
        (
            'v1',
            False,
        ),
        (
            'v1.3-r3',
            False,
        ),
        (
            'v1.3.3-r3',
            True,
        ),
        (
            'a.b.c',
            False,
        ),
        (
            'va.b.c',
            False,
        ),
        (
            '1.2.a',
            False,
        ),
        (
            'v1.3.3.3-r3',
            False,
        ),
        (
            '1.0.0+20130313144700',
            True,
        ),
        (
            '1.0.0-alpha+001',
            True,
        ),
    ]
)
def test_is_semantic_version(version, is_semantic):
    assert audeer.is_semantic_version(version) == is_semantic


@pytest.mark.parametrize(
    'uid, expected',
    [
        (audeer.uid(), True),
        (audeer.uid(short=True), True),
        (audeer.uid(from_string='from string'), True),
        (audeer.uid(from_string='from string', short=True), True),
        (audeer.uid(from_string='from string', short=True).upper(), True),
        (None, False),
        (1234, False),
        ('', False),
        ('some random string', False),
        (audeer.uid()[:-1], False),
        (audeer.uid(short=True)[:-1], False),
        ('00000000-0000-0000-0000-000000000000', True),
        ('000000000-0000-0000-0000-00000000000', False),
        ('?0000000-0000-0000-0000-000000000000', False),
    ]
)
def test_is_uid(uid, expected):
    assert audeer.is_uid(uid) == expected


def power(a: int = 0, *, b: int = 1):
    return a ** b


@pytest.mark.parametrize(
    'multiprocessing',
    [
        True, False,
    ]
)
@pytest.mark.parametrize(
    'num_workers',
    [
        1, 3, None,
    ]
)
@pytest.mark.parametrize(
    'task_fun, params',
    [
        (power, [([], {})]),
        (power, [([1], {})]),
        (power, [([1], {'b': 2})]),
        (power, [([], {'a': 1, 'b': 2})]),
        (power, [([x], {'b': x}) for x in range(5)]),
    ]
)
def test_run_tasks(multiprocessing, num_workers, task_fun, params):
    expected = [
        task_fun(*param[0], **param[1]) for param in params
    ]
    results = audeer.run_tasks(
        task_fun,
        params,
        num_workers=num_workers,
        multiprocessing=multiprocessing,
    )
    assert expected == results


@pytest.mark.parametrize('func,params,expected_output', [
    (
        lambda x, n: x ** n,
        [(2, n) for n in range(7)],
        [1, 2, 4, 8, 16, 32, 64],
    ),
    (3, None, [None]),
    (lambda x: x, ['hello'], ['hello']),
    (lambda x: x, ['hello', 1, print], ['hello', 1, print]),
])
def test_run_worker_threads(func, params, expected_output):
    num_workers = [None, 1, 2, 4, 5, 100]
    for n in num_workers:
        with pytest.warns(UserWarning):
            output = audeer.run_worker_threads(func, params, num_workers=n)
        assert output == expected_output


@pytest.mark.parametrize(
    'versions, expected_versions',
    [
        (
            ['1.0.0', '3.0.0', '1.10.1', '1.2.0', '1.0.1'],
            ['1.0.0', '1.0.1', '1.2.0', '1.10.1', '3.0.0'],
        ),
        (
            [
                '1.0.0-SNAPSHOT',
                '4.0.0-20200206.095424-2',
                '1.0.0',
                '2.0.0-20200131.102442-1',
                '3.0.0',
                '3.1.0',
                '4.0.0-20200206.095316-1',
                '3.2.0',
                '2.0.0-20200131.102728-2',
                '3.3.0',
                '3.4.0',
                '4.0.0-20200206.095534-3',
                '4.0.0',
            ],
            [
                '1.0.0',
                '1.0.0-SNAPSHOT',
                '2.0.0-20200131.102442-1',
                '2.0.0-20200131.102728-2',
                '3.0.0',
                '3.1.0',
                '3.2.0',
                '3.3.0',
                '3.4.0',
                '4.0.0',
                '4.0.0-20200206.095316-1',
                '4.0.0-20200206.095424-2',
                '4.0.0-20200206.095534-3',
            ],
        ),
        (
            ['v1.0.0', 'v1.0.1', 'v1.0.1-1-gdf29c4a'],
            ['v1.0.0', 'v1.0.1', 'v1.0.1-1-gdf29c4a'],
        ),
        # From https://github.com/postmarketOS/pmbootstrap/issues/342
        (
            ['22.7.3-r1.3', '22.7.3-r1'],
            ['22.7.3-r1', '22.7.3-r1.3'],
        ),
        (
            ['1.0.0', '1.1.1+1', '1.2.1', '1.2.0'],
            ['1.0.0', '1.1.1+1', '1.2.0', '1.2.1'],
        ),
    ]
)
def test_sort_versions(versions, expected_versions):
    sorted_versions = audeer.sort_versions(versions)
    assert sorted_versions == expected_versions


@pytest.mark.parametrize(
    'versions, error_message',
    [
        (
            ['1'],
            (
                "All version numbers have to be semantic versions, "
                "following 'X.Y.Z', "
                "where X, Y, Z are integers. "
                "But your version is: '1'."
            ),
        ),
    ]
)
def test_sort_versions_errors(versions, error_message):
    with pytest.raises(ValueError, match=error_message):
        audeer.sort_versions(versions)


@pytest.mark.parametrize(
    'input,expected_output',
    [
        (1, [1]),
        ('abc', ['abc']),
        (['abc', 1], ['abc', 1]),
        ((1, 2, 3), [1, 2, 3]),
        (len, [len]),
        ({1: 'a', 2: 'b'}, [1, 2]),
        ([], []),
        ('', ['']),
        (None, [None]),
    ],
)
def test_to_list(input, expected_output):
    assert audeer.to_list(input) == expected_output


@pytest.mark.parametrize(
    'short',
    [
        False,
        True,
    ]
)
@pytest.mark.parametrize(
    'from_string',
    [
        None,
        'example_string',
    ]
)
def test_uid(from_string, short):
    uid = audeer.uid(from_string=from_string, short=short)
    if short:
        assert len(uid) == 8
    else:
        assert len(uid) == 36
        for pos in [8, 13, 18, 23]:
            assert uid[pos] == '-'
        uid2 = audeer.uid(from_string=from_string, short=short)
        if from_string is not None:
            assert uid == uid2
        else:
            assert uid != uid2



import pytest


@pytest.fixture(autouse=True)
def docdir(request):
    # set temporal working directory in docstring examples
    tmpdir = request.getfixturevalue('tmpdir')
    with tmpdir.as_cwd():
        yield



import os
import platform
import stat
import time

import pytest

import audeer


@pytest.fixture(scope='function', autouse=False)
def tree(tmpdir, request):
    r"""Create file tree."""

    files = request.param
    paths = []

    for path in files:
        if os.name == 'nt':
            path = path.replace('/', os.path.sep)
        if path.endswith(os.path.sep):
            path = audeer.path(tmpdir, path)
            path = audeer.mkdir(path)
            path = path + os.path.sep
            paths.append(path)
        else:
            path = audeer.path(tmpdir, path)
            audeer.mkdir(os.path.dirname(path))
            path = audeer.touch(path)
            paths.append(path)

    yield paths


@pytest.mark.parametrize(
    'tree, root, files, archive_create, archive_extract, destination, '
    'expected',
    [
        (  # empty
            [],
            '.',
            [],
            'archive.zip',
            'archive.zip',
            '.',
            [],
        ),
        (
            [],
            '.',
            None,
            'archive.zip',
            'archive.zip',
            '.',
            [],
        ),
        (  # single file
            ['file.txt'],
            '.',
            'file.txt',
            'archive.zip',
            'archive.zip',
            '.',
            ['file.txt'],
        ),
        (  # archive folder does not exist
            ['file.txt'],
            '.',
            'file.txt',
            'sub/archive.zip',
            'sub/archive.zip',
            '.',
            ['file.txt'],
        ),
        (  # file in sub folder
            ['file.txt', 'sub/a/b/file.txt'],
            '.',
            ['sub/a/b/file.txt', 'file.txt'],
            'archive.zip',
            'archive.zip',
            '.',
            ['sub/a/b/file.txt', 'file.txt'],
        ),
        (
            ['file.txt', 'sub/a/b/file.txt'],
            '.',
            ['sub/a/b/file.txt'],
            'archive.zip',
            'archive.zip',
            '.',
            ['sub/a/b/file.txt'],
        ),
        (
            ['file.txt', 'sub/a/b/file.txt'],
            '.',
            ['file.txt'],
            'archive.zip',
            'archive.zip',
            '.',
            ['file.txt'],
        ),
        (  # include all files
            ['file.txt', 'sub/a/b/file.txt', '.hidden'],
            '.',
            None,
            'archive.zip',
            'archive.zip',
            '.',
            ['.hidden', 'file.txt', 'sub/a/b/file.txt'],
        ),
        (  # exclude empty folder
            ['file.txt', 'sub/a/b/file.txt', '.hidden', 'empty/'],
            '.',
            None,
            'archive.zip',
            'archive.zip',
            '.',
            ['.hidden', 'file.txt', 'sub/a/b/file.txt'],
        ),
        (  # tar.gz
            ['file.txt', 'sub/a/b/file.txt'],
            '.',
            ['sub/a/b/file.txt', 'file.txt'],
            'archive.tar.gz',
            'archive.tar.gz',
            '.',
            ['sub/a/b/file.txt', 'file.txt'],
        ),
        (  # root is sub folder
            ['sub/file.txt'],
            './sub',
            ['file.txt'],
            'archive.zip',
            'archive.zip',
            '.',
            ['file.txt'],
        ),
        (
            ['sub/file.txt'],
            './sub',
            None,
            'archive.zip',
            'archive.zip',
            '.',
            ['file.txt'],
        ),
        (  # destitation is sub folder
            ['file.txt'],
            '.',
            ['file.txt'],
            'archive.zip',
            'archive.zip',
            './sub',
            ['file.txt'],
        ),
        (  # relative path with ../
            ['sub/file.txt'],
            './sub',
            ['../sub/file.txt'],
            'archive.zip',
            'archive.zip',
            '.',
            ['file.txt'],
        ),
        pytest.param(  # root does not exit
            [],
            './sub',
            None,
            'archive.zip',
            'archive.zip',
            '.',
            None,
            marks=pytest.mark.xfail(raises=FileNotFoundError),
        ),
        pytest.param(  # root is not a directory
            ['file.txt'],
            'file.txt',
            None,
            'archive.zip',
            'archive.zip',
            '.',
            None,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(  # destination is not a directory
            ['file.txt'],
            '.',
            [],
            'archive.zip',
            'archive.zip',
            'file.txt',
            [],
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(  # archive to be extracted does not exit
            [],
            '.',
            [],
            'archive.zip',
            'bad.zip',
            '.',
            None,
            marks=pytest.mark.xfail(raises=FileNotFoundError),
        ),
        pytest.param(  # archive to be extracted is a directory
            ['sub/'],
            '.',
            [],
            'archive.zip',
            'sub',
            '.',
            None,
            marks=pytest.mark.xfail(raises=IsADirectoryError),
        ),
        pytest.param(  # file does not exit
            [],
            '.',
            ['file.txt'],
            'archive.zip',
            'archive.zip',
            '.',
            None,
            marks=pytest.mark.xfail(raises=FileNotFoundError),
        ),
        pytest.param(  # file not below root
            ['file.txt', 'sub/'],
            './sub',
            ['../file.txt'],
            'archive.zip',
            'archive.zip',
            '.',
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(  # archive type not supported
            [],
            '.',
            [],
            'archive.bad',
            'archive.zip',
            '.',
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(
            ['archive.bad'],
            '.',
            [],
            'archive.zip',
            'archive.bad',
            '.',
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(  # broken archive
            ['archive.zip'],
            '.',
            [],
            'archive.tar.gz',
            'archive.zip',
            '.',
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
    ],
    indirect=['tree'],
)
def test_archives(tmpdir, tree, root, files, archive_create,
                  archive_extract, destination, expected):

    root = audeer.path(tmpdir, root)
    destination = audeer.path(tmpdir, destination)
    archive_create = audeer.path(tmpdir, archive_create)
    archive_extract = audeer.path(tmpdir, archive_extract)

    if os.name == 'nt':
        if expected is not None:
            expected = [file.replace('/', os.path.sep) for file in expected]
        if isinstance(files, str):
            files = files.replace('/', os.path.sep)
        elif files is not None:
            files = [file.replace('/', os.path.sep) for file in files]

    # relative path

    audeer.create_archive(
        root,
        files,
        archive_create,
    )
    result = audeer.extract_archive(
        archive_extract,
        root,
        keep_archive=True,
    )
    assert result == expected
    for file in result:
        assert os.path.exists(audeer.path(root, file))
    assert os.path.exists(archive_extract)

    # list of archives

    result = audeer.extract_archives(
        [archive_extract, archive_extract],
        destination,
        keep_archive=True,
    )
    assert result == expected + expected
    for file in result:
        assert os.path.exists(audeer.path(root, file))
    assert os.path.exists(archive_extract)

    # absolute path

    if files is not None:
        if isinstance(files, str):
            files = audeer.path(root, files)
        else:
            files = [audeer.path(root, file) for file in files]
        audeer.create_archive(
            root,
            files,
            archive_create,
        )
        result = audeer.extract_archive(
            archive_extract,
            destination,
            keep_archive=True,
        )
        assert result == expected
        for file in result:
            assert os.path.exists(audeer.path(root, file))
        assert os.path.exists(archive_extract)

    # delete archive

    audeer.extract_archive(
        archive_extract,
        destination,
        keep_archive=False,
    )
    assert not os.path.exists(archive_extract)


@pytest.mark.parametrize('path,ext,basename', [
    ('~/.bashrc', None, '.bashrc'),
    ('file.tar.gz', None, 'file.tar'),
    ('/a/c.d/g', None, 'g'),
    (b'/a/c.d/g', None, 'g'),
    ('/a/c.d/g.exe', 'exe', 'g'),
    ('../.././README.md', '.md', 'README'),
    ('folder/file.txt', None, 'file'),
    ('folder/file.txt', 'txt', 'file'),
    (b'folder/file.txt', 'txt', 'file'),
])
def test_basename_wo_ext(path, ext, basename):
    b = audeer.basename_wo_ext(path, ext=ext)
    assert b == basename
    assert type(b) is str


@pytest.mark.parametrize('dirs,expected', [
    ([], ''),
    (
        [
            '/home/user/tmp/coverage/test',
            '/home/user/tmp/covert/operator',
            '/home/user/tmp/coven/members',
        ], '/home/user/tmp',
    ),
    (
        [
            '/home/user/tmp/coverage/test',
            '/home/user/tmp/covert/operator',
            '/home/user/tmp',
        ], '/home/user/tmp',
    ),
    (
        [
            '~/tmp/coverage/test',
            '~/tmp/covert/operator',
            '~/tmp/coven/members',
        ], f'{os.path.expanduser("~")}/tmp',
    ),
    (
        [
            '/etc/tmp/coverage/test',
            '/home/user/tmp/covert/operator',
            '/home/user/tmp/coven/members',
        ], '',
    ),
    (
        [
            '/home/user/tmp',
            '/home/user/tmp',
        ], '/home/user/tmp',
    ),
    (
        [
            '/home1/user/tmp',
            '/home2/user/tmp',
        ], '',
    ),
])
def test_common_directory(dirs, expected):
    common = audeer.common_directory(dirs)
    # Change paths always to Linux syntax
    _, common = os.path.splitdrive(common)
    _, expected = os.path.splitdrive(expected)
    common = common.replace('\\', '/')
    expected = expected.replace('\\', '/')
    # On MacOS we get a '/System/Volumes/Data' in front
    common = common.replace('/System/Volumes/Data', '')
    assert common == expected


def test_download_url(tmpdir):
    url = 'https://audeering.github.io/audeer/_static/favicon.png'
    audeer.download_url(url, tmpdir)
    audeer.download_url(url, tmpdir)
    dst = audeer.download_url(url, tmpdir, force_download=True)
    assert dst == os.path.join(tmpdir, os.path.basename(url))


@pytest.mark.parametrize('path,extension', [
    ('', ''),
    ('~/.bashrc', ''),
    ('file.tar.gz', 'gz'),
    ('/a/c.d/g', ''),
    ('/a/c.d/g.exe', 'exe'),
    ('../.././README.md', 'md'),
    (b'../.././README.md', 'md'),
    ('folder/file.txt', 'txt'),
    (b'folder/file.txt', 'txt'),
    ('test.WAV', 'WAV'),
    ('test.WaV', 'WaV'),
])
def test_file_extension(path, extension):
    ext = audeer.file_extension(path)
    assert ext == extension
    assert type(ext) is str


@pytest.mark.parametrize(
    'dir_list,expected,recursive,hidden',
    [
        ([], [], False, False),
        ([], [], True, False),
        (['a', 'b', 'c'], ['a', 'b', 'c'], False, False),
        (['a', 'b', 'c'], ['a', 'b', 'c'], True, False),
        (['a'], ['a'], False, False),
        (['a'], ['a'], True, False),
        (
            ['a', os.path.join('a', 'b'), os.path.join('a', 'b', 'c')],
            ['a', os.path.join('a', 'b'), os.path.join('a', 'b', 'c')],
            True,
            False,
        ),
        # hidden
        (['a', '.b'], ['a'], True, False),
        (['a', '.b'], ['.b', 'a'], True, True),
        (
            ['a', '.b', os.path.join('a', '.b'), os.path.join('a', '.b', 'c')],
            ['a'],
            True,
            False,
        ),
        (
            ['a', '.b', os.path.join('a', '.b'), os.path.join('a', '.b', 'c')],
            ['.b', 'a', os.path.join('a', '.b'), os.path.join('a', '.b', 'c')],
            True,
            True,
        ),
    ],
)
def test_list_dir_names(tmpdir, dir_list, expected, recursive, hidden):

    dir_tmp = tmpdir.mkdir('folder')
    directories = []
    for directory in dir_list:
        directory = os.path.join(str(dir_tmp), directory)
        directories.append(audeer.mkdir(directory))

    for directory in directories:
        assert os.path.isdir(directory)

    path = os.path.join(str(dir_tmp), '.')
    dirs = audeer.list_dir_names(
        path,
        basenames=False,
        recursive=recursive,
        hidden=hidden,
    )
    assert dirs == [audeer.path(dir_tmp, d) for d in expected]
    assert type(dirs) is list

    # test basenames
    dirs = audeer.list_dir_names(
        path,
        basenames=True,
        recursive=recursive,
        hidden=hidden,
    )
    assert dirs == expected


def test_list_dir_names_errors(tmpdir):
    with pytest.raises(NotADirectoryError):
        file = audeer.touch(audeer.path(tmpdir, 'file.txt'))
        audeer.list_dir_names(file)
    with pytest.raises(FileNotFoundError):
        audeer.list_dir_names('not-existent')


@pytest.mark.parametrize(
    'files,path,filetype,expected,recursive,hidden',
    [
        # empty
        ([], '.', '', [], False, False),
        ([], '.', '', [], True, False),
        ([], '.', 'wav', [], False, False),
        ([], '.', 'wav', [], True, False),
        # file
        (
            ['file.txt'],
            'file.txt',
            '',
            ['file.txt'],
            False,
            False,
        ),
        pytest.param(
            [os.path.join('sub', 'file.txt')],
            'file.txt',
            '',
            [],
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        (
            [os.path.join('sub', 'file.txt')],
            'file.txt',
            '',
            [os.path.join('sub', 'file.txt')],
            True,
            False,
        ),
        pytest.param(
            [],
            'file',
            '',
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        (
            [
                't1.wav',
                't2.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't2.wav'),
            ],
            't1.wav',
            '',
            [os.path.join('sub', 't1.wav'), 't1.wav'],
            True,
            False,
        ),
        # folder
        (
            ['t3.ogg', 't2.wav', 't1.wav'],
            '.',
            '',
            ['t1.wav', 't2.wav', 't3.ogg'],
            False,
            False,
        ),
        (
            [
                't3.ogg',
                't2.wav',
                't1.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't1.ogg'),
            ],
            '.',
            '',
            ['t1.wav', 't2.wav', 't3.ogg'],
            False,
            False,
        ),
        (
            [
                't3.ogg',
                't2.wav',
                't1.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 'sub', 't1.ogg'),
            ],
            '.',
            '',
            [
                os.path.join('sub', 'sub', 't1.ogg'),
                os.path.join('sub', 't1.wav'),
                't1.wav',
                't2.wav',
                't3.ogg',
            ],
            True,
            False,
        ),
        pytest.param(
            [],
            'does-not-exist',
            '',
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(
            [],
            os.path.join('does', 'not', 'exist'),
            '',
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        # filetype
        (
            [
                't3.ogg',
                't2.wav',
                't1.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 'sub', 't1.ogg'),
            ],
            '.',
            'ogg',
            ['t3.ogg'],
            False,
            False,
        ),
        (
            [
                't3.ogg',
                't2.wav',
                't1.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 'sub', 't1.ogg'),
            ],
            '.',
            'ogg',
            [
                os.path.join('sub', 'sub', 't1.ogg'),
                't3.ogg',
            ],
            True,
            False,
        ),
        (
            [
                't1.wav',
                os.path.join('sub', 't1.wav'),
            ],
            't1.wav',
            '',
            ['t1.wav'],
            False,
            False,
        ),
        # pattern
        (
            [
                't1.wav',
                't2.ogg',
                's3.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't2.ogg'),
                os.path.join('sub', 's3.wav'),
            ],
            't*',
            '',
            [
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't2.ogg'),
                't1.wav',
                't2.ogg'
            ],
            True,
            False,
        ),
        (
            [
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't2.ogg'),
                os.path.join('sub', 's3.wav'),
            ],
            't*',
            '',
            [
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't2.ogg'),
            ],
            True,
            False,
        ),
        (
            [
                't1.wav',
                't2.ogg',
                's3.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't2.ogg'),
                os.path.join('sub', 's3.wav'),
            ],
            'x*',
            '',
            [],
            True,
            False,
        ),
        (
            [
                't1.wav',
                't2.wav',
                's1.wav',
                's2.wav',
            ],
            't?.wav',
            '',
            ['t1.wav', 't2.wav'],
            False,
            False,
        ),
        (
            [
                't1.wav',
                't2.wav',
                's1.wav',
                's2.wav',
            ],
            '[ts]1.wav',
            '',
            ['s1.wav', 't1.wav'],
            False,
            False,
        ),
        (
            [
                't1.wav',
                't2.wav',
                's1.wav',
                's.wav',
            ],
            '[ts]?.wav',
            '',
            ['s1.wav', 't1.wav', 't2.wav'],
            False,
            False,
        ),
        (
            [
                't1.wav',
                't2.wav',
                's1.wav',
                's.wav',
                'x.wav',
            ],
            '[ts]*.wav',
            '',
            ['s.wav', 's1.wav', 't1.wav', 't2.wav'],
            False,
            False,
        ),
        (
            [],
            'file*',
            '',
            [],
            False,
            False,
        ),
        (
            [],
            '?ile',
            '',
            [],
            False,
            False,
        ),
        pytest.param(
            [],
            os.path.join('does', 'not', 'exist', 'file*'),
            '',
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        pytest.param(
            [],
            os.path.join('not!a[pattern'),
            '',
            None,
            False,
            False,
            marks=pytest.mark.xfail(raises=NotADirectoryError),
        ),
        # pattern + filetype
        (
            [
                't1.wav',
                't2.ogg',
                's3.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', 't2.ogg'),
                os.path.join('sub', 's3.wav'),
            ],
            't*',
            'ogg',
            [
                os.path.join('sub', 't2.ogg'),
                't2.ogg'
            ],
            True,
            False,
        ),
        # hidden
        (
            ['.file.txt'],
            '.file.txt',
            '',
            [],
            False,
            False,
        ),
        (
            ['.file.txt'],
            '.file.txt',
            '',
            ['.file.txt'],
            False,
            True,
        ),
        (
            [os.path.join('sub', '.file.txt')],
            '.file.txt',
            '',
            [],
            True,
            False,
        ),
        (
            [os.path.join('sub', '.file.txt')],
            '.file.txt',
            '',
            [os.path.join('sub', '.file.txt')],
            True,
            True,
        ),
        (
            [
                't1.wav',
                '.t2.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', '.t2.wav'),
            ],
            '',
            '',
            [
                't1.wav',
            ],
            False,
            False,
        ),
        (
            [
                't1.wav',
                '.t2.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', '.t2.wav'),
            ],
            '',
            '',
            [
                os.path.join('sub', 't1.wav'),
                't1.wav',
            ],
            True,
            False,
        ),
        (
            [
                't1.wav',
                '.t2.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', '.t2.wav'),
            ],
            '',
            '',
            [
                '.t2.wav',
                't1.wav',
            ],
            False,
            True,
        ),
        (
            [
                't1.wav',
                '.t2.wav',
                os.path.join('sub', 't1.wav'),
                os.path.join('sub', '.t2.wav'),
            ],
            '',
            '',
            [
                '.t2.wav',
                os.path.join('sub', '.t2.wav'),
                os.path.join('sub', 't1.wav'),
                't1.wav',
            ],
            True,
            True,
        ),
    ],
)
def test_list_file_names(tmpdir, files, path, filetype, expected,
                         recursive, hidden):
    dir_tmp = tmpdir.mkdir('folder')
    dir_tmp.mkdir('subfolder')
    path = os.path.join(str(dir_tmp), path)
    for file in files:
        # Create the files
        file_tmp = dir_tmp.join(file)
        audeer.mkdir(os.path.dirname(file_tmp))
        file_tmp.write('')
    f = audeer.list_file_names(
        path,
        filetype=filetype,
        basenames=False,
        recursive=recursive,
        hidden=hidden,
    )
    # test full path
    assert f == [audeer.path(dir_tmp, f) for f in expected]
    assert type(f) is list
    # test basenames
    f = audeer.list_file_names(
        path,
        filetype=filetype,
        basenames=True,
        recursive=recursive,
        hidden=hidden,
    )
    assert f == expected


def test_md5_errors():
    with pytest.raises(FileNotFoundError):
        audeer.md5('does/not/exist')


@pytest.mark.parametrize(
    'file, content, expected',
    [
        (  # empty file
            'file.txt',
            None,
            'd41d8cd98f00b204e9800998ecf8427e',
        ),
        (  # different content
            'file.txt',
            'hello world',
            '5eb63bbbe01eeed093cb22bb8f5acdc3',
        ),
        (
            'file.txt',
            'Hello World',
            'b10a8db164e0754105b7a99be72e3fe5',
        ),
        (  # different filename
            'file.TXT',
            'Hello World',
            'b10a8db164e0754105b7a99be72e3fe5',
        ),
    ],
)
def test_md5_file(tmpdir, file, content, expected):

    path = audeer.path(tmpdir, file)
    with open(path, 'w') as fp:
        if content is not None:
            fp.write(content)

    assert audeer.md5(path) == expected


@pytest.mark.parametrize(
    'tree, content, expected',
    [
        (  # empty folder
            [],
            None,
            'd41d8cd98f00b204e9800998ecf8427e',
        ),
        (  # folder with different content
            ['f'],
            None,
            '8fa14cdd754f91cc6554c9e71929cce7',
        ),
        (
            ['sub/f'],
            None,
            '1af042d5a4ec129583f6093f98f64118',
        ),
        (
            ['f', 'sub/f'],
            None,
            'b540f38948f445622adc657a757f4b0d',
        ),
        (
            ['f', 'sub/g'],
            None,
            '305107efbb15f9334d22ae4fbeec4de6',
        ),
        (
            ['f', 'sub/g'],
            'hello world',
            '47829eb8ef287d0d72e0fed9b96d258d',
        ),
        (
            ['f', 'sub/g'],
            'Hello World',
            '442d96d7c43bb18f247888408e5d6977',
        ),
        (  # with empty sub folder
            ['f', 'sub/g', 'sub/'],
            None,
            '305107efbb15f9334d22ae4fbeec4de6',
        ),
        (  # with hidden file
            ['f', 'sub/g', '.hidden'],
            None,
            '97490b233a7717aec19023e28443a1bf',
        ),
        (  # umlaute
            ['ä', 'ö', 'ü', 'ß'],
            None,
            '622165ad36122984c6b2c7ba466aa262',
        ),
    ],
    indirect=['tree'],
)
def test_md5_folder(tmpdir, tree, content, expected):

    if content is not None:
        for path in tree:
            with open(path, 'w') as fp:
                fp.write(content)

    assert audeer.md5(tmpdir) == expected


def test_mkdir(tmpdir):
    # New dir
    path = str(tmpdir.mkdir('folder1'))
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path
    # Existing dir
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path
    # Existing dir with content
    dir_tmp = tmpdir.mkdir('folder2')
    f = dir_tmp.join('file.txt')
    f.write('')
    path = str(dir_tmp)
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path
    # Relative path
    path = str(tmpdir.mkdir('folder3'))
    current_path = os.getcwd()
    os.chdir(path)
    p = audeer.mkdir('folder4')
    os.chdir(current_path)
    assert os.path.isdir(p) is True
    assert p == os.path.join(path, 'folder4')
    # Subdirectories
    os.chdir(path)
    p = audeer.mkdir('folder5/folder6')
    os.chdir(current_path)
    assert os.path.isdir(p) is True
    assert p == os.path.join(path, 'folder5', 'folder6')
    # Path in bytes
    path = str(tmpdir.mkdir('folder7'))
    path = bytes(path, 'utf8')
    p = audeer.mkdir(path)
    assert os.path.isdir(p) is True
    assert p == path.decode('utf8')
    # Empty dir
    path = ''
    p = audeer.mkdir(path)
    assert p == path
    # Mode, see https://stackoverflow.com/a/705088
    # Default mode
    path = os.path.join(str(tmpdir.mkdir('folder8')), 'sub-folder')
    os.umask(0)
    p = audeer.mkdir(path)
    mode = stat.S_IMODE(os.stat(p).st_mode)
    expected_mode = int('777', 8)
    assert mode == expected_mode
    # Non-default modes
    # Under Windows, changing permissions does not work,
    # there we always expect 777
    path = os.path.join(str(tmpdir.mkdir('folder9')), 'sub-folder')
    os.umask(0)
    p = audeer.mkdir(path, mode=0o775)
    expected_mode = '775'
    if platform.system() == 'Windows':
        expected_mode = '777'
    mode = stat.S_IMODE(os.stat(p).st_mode)
    assert mode == int(expected_mode, 8)
    assert mode != int('755', 8)
    path = os.path.join(str(tmpdir.mkdir('folder10')), 'sub-folder')
    os.umask(0)
    p = audeer.mkdir(path, mode=0o755)
    expected_mode = '755'
    if platform.system() == 'Windows':
        expected_mode = '777'
    mode = stat.S_IMODE(os.stat(p).st_mode)
    assert mode == int(expected_mode, 8)
    assert mode != int('775', 8)


@pytest.mark.parametrize(
    'src_file, dst_file',
    [
        (
            'file1',
            'file1',
        ),
        (
            'file1',
            'file2',
        ),
    ]
)
def test_move_file(tmpdir, src_file, dst_file):

    tmp_path = str(tmpdir.mkdir('folder'))
    tmp_path = audeer.mkdir(tmp_path)

    src_path = audeer.touch(os.path.join(tmp_path, src_file))
    dst_path = os.path.join(tmp_path, dst_file)

    audeer.move_file(src_path, dst_path)

    if src_file != dst_file:
        assert not os.path.exists(src_path)
    assert os.path.exists(dst_path)


@pytest.mark.parametrize(
    'path, new_extension, ext, expected_path',
    [
        ('', '', None, ''),
        ('', 'txt', None, ''),
        ('', '', 'rst', ''),
        ('', 'txt', 'rst', ''),
        ('file', '', None, 'file'),
        ('file', 'txt', None, 'file.txt'),
        ('file.txt', 'wav', None, 'file.wav'),
        ('test/file.txt', 'wav', None, 'test/file.wav'),
        ('a/b/../file.txt', 'wav', None, 'a/b/../file.wav'),
        ('file.txt', 'wav', 'txt', 'file.wav'),
        ('file.txt', 'wav', '.txt', 'file.wav'),
        ('file.txt', '.wav', 'txt', 'file.wav'),
        ('file.txt', '.wav', '.txt', 'file.wav'),
        ('file.a.b', 'wav', 'a.b', 'file.wav'),
        ('file.a.b', 'wav', '.a.b', 'file.wav'),
        ('file', 'wav', 'ext', 'file'),
        ('file.txt', 'wav', 'ext', 'file.txt'),
        ('file.txt', 'wav', 't', 'file.txt'),
    ]
)
def test_replace_file_extension(path, new_extension, ext, expected_path):
    new_path = audeer.replace_file_extension(path, new_extension, ext=ext)
    assert new_path == expected_path


def test_rmdir(tmpdir):
    # Non existing dir
    audeer.rmdir('non-esitent')
    # Folder with file content
    dir_tmp = tmpdir.mkdir('folder')
    f = dir_tmp.join('file.txt')
    f.write('')
    path = str(dir_tmp)
    p = audeer.mkdir(path)
    with pytest.raises(NotADirectoryError):
        audeer.rmdir(os.path.join(p, 'file.txt'))
    audeer.rmdir(p)
    assert not os.path.exists(p)
    # Folder with folder content
    path = str(tmpdir.mkdir('folder'))
    p = audeer.mkdir(os.path.join(path, 'sub-folder'))
    audeer.rmdir(os.path.dirname(p))
    assert not os.path.exists(p)
    assert not os.path.exists(os.path.dirname(p))
    # Relative path
    path = str(tmpdir.mkdir('folder'))
    current_path = os.getcwd()
    os.chdir(os.path.dirname(path))
    assert os.path.exists(path)
    audeer.rmdir('folder')
    assert not os.path.exists(path)
    os.chdir(current_path)


def test_touch(tmpdir):
    path = str(tmpdir.mkdir('folder1'))
    path = audeer.mkdir(path)
    path = os.path.join(path, 'file')
    assert not os.path.exists(path)
    audeer.touch(path)
    assert os.path.exists(path)
    stat = os.stat(path)
    time.sleep(.1)
    audeer.touch(path)
    assert os.path.exists(path)
    new_stat = os.stat(path)
    assert stat.st_atime < new_stat.st_atime
    assert stat.st_mtime < new_stat.st_mtime



import errno
import fnmatch
import hashlib
import itertools
import os
import platform
import shutil
import tarfile
import typing
import urllib.request
import zipfile

from audeer.core.path import path as safe_path
from audeer.core.tqdm import (
    format_display_message,
    progress_bar,
)
from audeer.core.utils import to_list


# Exclude common_directory example from doctest
# on Windows and MacOS
# (which adds /System/Volumes/Data in front in the Github runner)
# as it outputs a path in Linux syntax in the example
if platform.system() in ['Darwin', 'Windows']:  # pragma: no cover
    __doctest_skip__ = [
        'common_directory',
        'list_dir_names',
        'list_file_names',
    ]


def basename_wo_ext(
        path: typing.Union[str, bytes],
        *,
        ext: str = None
) -> str:
    """File basename without file extension.

    Args:
        path: file or directory name
        ext: explicit extension to be removed

    Returns:
        basename of directory or file without extension

    Examples:

        >>> path = '/test/file.wav'
        >>> basename_wo_ext(path)
        'file'

    """
    path = safe_path(path)
    path = os.path.basename(path)
    if ext is not None:
        if not ext.startswith('.'):
            ext = '.' + ext  # 'mp3' => '.mp3'
        if path.endswith(ext):
            path = path[:-len(ext)]
    else:
        path = os.path.splitext(path)[0]
    return path


def common_directory(
    dirs: typing.Sequence[str],
    *,
    sep: str = os.path.sep,
) -> str:
    r"""Find common directory path.

    Args:
        dirs: list of directories
        sep: path separator

    Returns:
        part of the directory tree that is common to all the directories

    Examples:
        >>> paths = [
        ...     '/home/user1/tmp/coverage/test',
        ...     '/home/user1/tmp/covert/operator',
        ...     '/home/user1/tmp/coven/members',
        ... ]
        >>> common_directory(paths)
        '/home/user1/tmp'

    """
    def all_names_equal(name):
        return all(n == name[0] for n in name[1:])

    dirs = [safe_path(path) for path in dirs]
    by_directory_levels = zip(*[p.split(sep) for p in dirs])
    return sep.join(x[0] for x in itertools.takewhile(
        all_names_equal, by_directory_levels,
    ))


def create_archive(
        root: str,
        files: typing.Optional[typing.Union[str, typing.Sequence[str]]],
        archive: str,
        *,
        verbose: bool = False,
):
    r"""Create ZIP or TAR.GZ archive.

    If a list with ``files`` is provided,
    only those files will be included to the archive.
    In that case the files are added in the given order,
    which may have an influence on the checksum of the archive.

    If ``files`` is set to ``None``,
    all files below ``root`` will be added in sorted order.
    This includes hidden files and files from sub-folders,
    but not empty folders.

    Args:
        root: path to root folder of archive.
            Only files below ``root`` can be included
            and will be stored relative to ``root``
        files: files that will be included in the archive.
            Absolute and relative file paths are possible,
            as long as the files are below ``root``.
            If set to ``None``
            all files below ``root``
            will be added to the archive
        archive: path to archive file.
            The type of the archive
            is determined from its file extension
        verbose: if ``True`` a progress bar is shown

    Raises:
        FileNotFoundError: if ``root`` or a file in ``files`` is not found
        NotADirectoryError: if ``root`` is not a directory
        RuntimeError: if archive does not end with ``zip`` or ``tar.gz``
            or a file in ``files`` is not below ``root``

    Examples:
        >>> _ = touch('a.txt')
        >>> _ = touch('b.txt')
        >>> create_archive('.', None, 'archive.zip')
        >>> extract_archive('archive.zip', '.')
        ['a.txt', 'b.txt']
        >>> create_archive('.', ['a.txt'], 'archive.tar.gz')
        >>> extract_archive('archive.tar.gz', '.')
        ['a.txt']

    """
    root = safe_path(root)
    archive = safe_path(archive)
    mkdir(os.path.dirname(archive))

    if not os.path.exists(root):
        raise FileNotFoundError(
            errno.ENOENT,
            os.strerror(errno.ENOENT),
            root,
        )

    if not os.path.isdir(root):
        raise NotADirectoryError(
            errno.ENOTDIR,
            os.strerror(errno.ENOTDIR),
            root,
        )

    if files is None:

        files = list_file_names(
            root,
            basenames=True,
            recursive=True,
            hidden=True,
        )

    else:

        files_org = to_list(files)
        files = []

        for file in files_org:

            # convert to absolute path
            if not os.path.isabs(file):
                path = safe_path(root, file)
            else:
                path = safe_path(file)

            # file is below root
            if not path.startswith(root):
                raise RuntimeError(
                    f"Only files below "
                    f"'{root}' "
                    f"can be included. "
                    f"This is not the case with "
                    f"'{file}'"
                )

            # file exists
            if not os.path.exists(path):
                raise FileNotFoundError(
                    errno.ENOENT,
                    os.strerror(errno.ENOENT),
                    file,
                )

            # convert to relative path
            files.append(path[len(root) + 1:])

    # Progress bar arguments
    desc = format_display_message(
        f'Create {os.path.basename(archive)}',
        pbar=True,
    )
    disable = not verbose

    if archive.endswith('zip'):
        with zipfile.ZipFile(archive, 'w', zipfile.ZIP_DEFLATED) as zf:
            for file in progress_bar(files, desc=desc, disable=disable):
                full_file = safe_path(root, file)
                zf.write(full_file, arcname=file)
    elif archive.endswith('tar.gz'):
        with tarfile.open(archive, 'w:gz') as tf:
            for file in progress_bar(files, desc=desc, disable=disable):
                full_file = safe_path(root, file)
                tf.add(full_file, file)
    else:
        raise RuntimeError(
            f'You can only create a ZIP or TAR.GZ archive, '
            f'not {archive}'
        )


def download_url(
        url: str,
        destination: str,
        *,
        force_download: bool = False,
        verbose: bool = False,
) -> str:
    r"""Download URL to destination.

    Args:
        url: URL of file to download
        destination: file or folder to store file locally
        force_download: if ``True`` forces the artifact to be downloaded
            even if it exists locally already
        verbose: if ``True`` a progress bar is shown

    Returns:
        path of locally stored file

    Examples:
        >>> dst = download_url('https://audeering.github.io/audeer/_static/favicon.png', '.')
        >>> os.path.basename(dst)
        'favicon.png'

    """  # noqa: E501
    destination = safe_path(destination)
    if os.path.isdir(destination):
        destination = os.path.join(destination, os.path.basename(url))
    if os.path.exists(destination) and not force_download:
        return destination

    with progress_bar(
            disable=not verbose,
            desc=format_display_message(f'Downloading {url}', pbar=True),
    ) as pbar:

        def bar_update(block_num, block_size, total_size):
            if pbar.total is None and total_size:
                pbar.total = total_size
            pbar.update(block_size)

        urllib.request.urlretrieve(url, destination, reporthook=bar_update)

    return destination


def extract_archive(
        archive: str,
        destination: str,
        *,
        keep_archive: bool = True,
        verbose: bool = False,
) -> typing.List[str]:
    r"""Extract ZIP or TAR.GZ file.

    Args:
        archive: path to ZIP or TAR.GZ file
        destination: folder where the files will be extracted.
            If the folder does not exists,
            it will be created
        keep_archive: if ``False`` delete archive file after extraction
        verbose: if ``True`` a progress bar is shown

    Returns:
        paths of extracted files relative to ``destintation``
        in order they were added to the archive

    Raises:
        FileNotFoundError: if ``archive`` is not found
        IsADirectoryError: if ``archive`` is a directory
        NotADirectoryError: if ``destination`` is not a directory
        RuntimeError: if ``archive`` is not a ZIP or TAR.GZ file
        RuntimeError: if ``archive`` is malformed

    Examples:
        >>> _ = touch('a.txt')
        >>> create_archive('.', None, 'archive.zip')
        >>> extract_archive('archive.zip', '.')
        ['a.txt']
        >>> extract_archive('archive.zip', 'sub')
        ['a.txt']

    """
    archive = safe_path(archive)
    destination = safe_path(destination)

    if not os.path.exists(archive):
        raise FileNotFoundError(
            errno.ENOENT,
            os.strerror(errno.ENOENT),
            archive,
        )

    if os.path.isdir(archive):
        raise IsADirectoryError(
            errno.EISDIR,
            os.strerror(errno.EISDIR),
            archive,
        )

    if os.path.exists(destination):
        destination_created = False
    else:
        mkdir(destination)
        destination_created = True

    if not os.path.isdir(destination):
        raise NotADirectoryError(
            errno.ENOTDIR,
            os.strerror(errno.ENOTDIR),
            destination,
        )

    # Progress bar arguments
    desc = format_display_message(
        f'Extract {os.path.basename(archive)}',
        pbar=True,
    )
    disable = not verbose

    try:
        if archive.endswith('zip'):
            with zipfile.ZipFile(archive, 'r') as zf:
                members = zf.infolist()
                for member in progress_bar(
                    members,
                    desc=desc,
                    disable=disable,
                ):
                    zf.extract(member, destination)
                files = [m.filename for m in members]
        elif archive.endswith('tar.gz'):
            with tarfile.open(archive, 'r') as tf:
                members = tf.getmembers()
                for member in progress_bar(
                    members,
                    desc=desc,
                    disable=disable,
                ):
                    tf.extract(member, destination, numeric_owner=True)
                files = [m.name for m in members]
        else:
            raise RuntimeError(
                f'You can only extract ZIP and TAR.GZ files, '
                f'not {archive}'
            )
    except (EOFError, zipfile.BadZipFile, tarfile.ReadError):
        raise RuntimeError(f'Broken archive: {archive}')
    except (KeyboardInterrupt, Exception):  # pragma: no cover
        # Clean up broken extraction files
        if destination_created:
            if os.path.exists(destination):
                shutil.rmtree(destination)
        raise

    if not keep_archive:
        os.remove(archive)

    if os.name == 'nt':  # pragma: no cover
        # replace '/' with '\' on Windows
        files = [file.replace('/', os.path.sep) for file in files]

    return files


def extract_archives(
        archives: typing.Sequence[str],
        destination: str,
        *,
        keep_archive: bool = True,
        verbose: bool = False,
) -> typing.List[str]:
    r"""Extract multiple ZIP or TAR.GZ archives at once.

    Args:
        archives: paths of ZIP or TAR.GZ files
        destination: folder where the files will be extracted.
            If the folder does not exists,
            it will be created
        keep_archive: if ``False`` delete archive files after extraction
        verbose: if ``True`` a progress bar is shown

    Returns:
        paths of extracted files relative to ``destintation``
        in order they were added to the archives

    Raises:
        FileNotFoundError: if an archive is not found
        IsADirectoryError: if an archive is a directory
        NotADirectoryError: if ``destination`` is not a directory
        RuntimeError: if an archive is not a ZIP or TAR.GZ file
        RuntimeError: if an archive file is malformed

    Examples:
        >>> _ = touch('a.txt')
        >>> create_archive('.', ['a.txt'], 'archive.zip')
        >>> _ = touch('b.txt')
        >>> create_archive('.', ['b.txt'], 'archive.tar.gz')
        >>> extract_archives(['archive.zip', 'archive.tar.gz'], '.')
        ['a.txt', 'b.txt']

    """
    with progress_bar(
        total=len(archives),
        disable=not verbose,
    ) as pbar:
        member_names = []
        for archive in archives:
            desc = format_display_message(
                f'Extract {os.path.basename(archive)}',
                pbar=True,
            )
            pbar.set_description_str(desc)
            pbar.refresh()
            member_names += extract_archive(
                archive,
                destination,
                keep_archive=keep_archive,
                verbose=False,
            )
            pbar.update()

    return member_names


def file_extension(
        path: typing.Union[str, bytes]
) -> str:
    """File extension.

    Args:
        path: path to file

    Returns:
        extension of file without "."

    Examples:
        >>> path = '/test/file.wav'
        >>> file_extension(path)
        'wav'

    """
    path = safe_path(path)
    return os.path.splitext(path)[-1][1:]


def list_dir_names(
        path: typing.Union[str, bytes],
        *,
        basenames: bool = False,
        recursive: bool = False,
        hidden: bool = True,
) -> typing.List[str]:
    """List of folder names located inside provided path.

    Args:
        path: path to directory
        basenames: if ``True`` return relative path in respect to ``path``
        recursive: if ``True`` includes subdirectories
        hidden: if ``True`` includes directories starting with a dot (``.``)

    Returns:
        list of paths to directories

    Raises:
        NotADirectoryError: if path is not a directory
        FileNotFoundError: if path does not exists

    Examples:
        >>> _ = mkdir('path/a/.b/c')
        >>> list_dir_names(
        ...     'path',
        ...     basenames=True,
        ... )
        ['a']
        >>> list_dir_names(
        ...     'path',
        ...     basenames=True,
        ...     recursive=True,
        ... )
        ['a', 'a/.b', 'a/.b/c']
        >>> list_dir_names(
        ...     'path',
        ...     basenames=True,
        ...     recursive=True,
        ...     hidden=False,
        ... )
        ['a']

    """
    path = safe_path(path)

    def helper(p: str, paths: typing.List[str]):
        ps = [os.path.join(p, x) for x in os.listdir(p)]
        ps = [x for x in ps if os.path.isdir(x)]
        if not hidden:
            ps = [x for x in ps if not os.path.basename(x).startswith('.')]
        paths.extend(ps)
        if len(ps) > 0 and recursive:
            for p in ps:
                helper(p, paths)

    paths = []
    helper(path, paths)
    if basenames:
        paths = [p[len(path) + 1:] for p in paths]

    return sorted(paths)


def list_file_names(
        path: typing.Union[str, bytes],
        *,
        filetype: str = '',
        basenames: bool = False,
        recursive: bool = False,
        hidden: bool = False,
) -> typing.List[str]:
    """List of file names inferred from provided path.

    Args:
        path: path to directory,
            or path to directory plus file matching pattern,
            e.g. ``'dir/file.txt'``.
            If ``recursive`` is ``True``
            returns all files named ``file.txt``
            from all sub-folders.
            Besides the filename ``*``, ``?``, ``[seq]``,
            and ``[!seq]`` can be used as pattern,
            compare :mod:`fnmatch`
        filetype: optional consider only this filetype
        basenames: if ``True`` return relative path in respect to ``path``
        recursive: if ``True`` includes subdirectories
        hidden: if ``True`` includes files starting with a dot (``.``)

    Returns:
        alphabetically sorted list of path(s) to file(s)

    Raises:
        NotADirectoryError: if ``path`` is a directory or file
            that does not exist,
            or ``path`` is a pattern
            and ``os.dirname(path)`` does not exist

    Examples:
        >>> dir_path = mkdir('path')
        >>> _ = touch(os.path.join(dir_path, 'file.wav'))
        >>> _ = touch(os.path.join(dir_path, 'File.wav'))
        >>> _ = touch(os.path.join(dir_path, '.lock'))
        >>> sub_dir_path = mkdir(os.path.join('path', 'sub'))
        >>> _ = touch(os.path.join(sub_dir_path, 'file.ogg'))
        >>> _ = touch(os.path.join(sub_dir_path, '.lock'))
        >>> list_file_names(
        ...     dir_path,
        ...     basenames=True,
        ... )
        ['File.wav', 'file.wav']
        >>> list_file_names(
        ...     dir_path,
        ...     basenames=True,
        ...     hidden=True,
        ... )
        ['.lock', 'File.wav', 'file.wav']
        >>> list_file_names(
        ...     dir_path,
        ...     basenames=True,
        ...     recursive=True,
        ... )
        ['File.wav', 'file.wav', 'sub/file.ogg']
        >>> list_file_names(
        ...     dir_path,
        ...     basenames=True,
        ...     recursive=True,
        ...     hidden=True,
        ... )
        ['.lock', 'File.wav', 'file.wav', 'sub/.lock', 'sub/file.ogg']
        >>> list_file_names(
        ...     os.path.join(dir_path, 'f*'),
        ...     basenames=True,
        ...     recursive=True,
        ...     hidden=True,
        ... )
        ['file.wav', 'sub/file.ogg']
        >>> list_file_names(
        ...     os.path.join(dir_path, '[fF]*'),
        ...     basenames=True,
        ...     recursive=True,
        ...     hidden=True,
        ... )
        ['File.wav', 'file.wav', 'sub/file.ogg']
        >>> list_file_names(
        ...     os.path.join(dir_path, '[!f]*'),
        ...     basenames=True,
        ...     recursive=True,
        ...     hidden=True,
        ... )
        ['.lock', 'File.wav', 'sub/.lock']
        >>> list_file_names(
        ...     os.path.join(dir_path, 'f*'),
        ...     filetype='ogg',
        ...     basenames=True,
        ...     recursive=True,
        ...     hidden=True,
        ... )
        ['sub/file.ogg']

    """
    path = safe_path(path)

    if os.path.isdir(path):

        pattern = None
        folder = path

    elif os.path.exists(path) and not recursive:

        if not hidden and os.path.basename(path).startswith('.'):
            return []
        if basenames:
            path = os.path.basename(path)
        return [path]

    else:

        pattern = os.path.basename(path)
        folder = os.path.dirname(path)

        if not os.path.isdir(folder):
            raise NotADirectoryError(folder)

    def helper(p: str, paths: typing.List[str]):
        ps = [os.path.join(p, x) for x in os.listdir(p)]
        folders = [x for x in ps if os.path.isdir(x)]
        files = [x for x in ps if os.path.isfile(x)]
        if pattern:
            files = [
                file for file in files
                if fnmatch.fnmatch(os.path.basename(file), f'{pattern}')
            ]
        if filetype:
            files = [
                file for file in files
                if fnmatch.fnmatch(os.path.basename(file), f'*{filetype}')
            ]
        paths.extend(files)
        if len(folders) > 0 and recursive:
            for p in folders:
                helper(p, paths)

    paths = []
    helper(folder, paths)

    def is_pattern(pattern):
        return (
                '*' in pattern or
                '?' in pattern or
                ('[' in pattern and ']' in pattern)
        )

    # if we have no match,
    # raise an error unless
    # 1. path is a folder (i.e. pattern is None)
    # 2. or we have a valid pattern
    if (
            len(paths) == 0
            and pattern is not None
            and not is_pattern(pattern)
    ):
        raise NotADirectoryError(path)

    if not hidden:
        paths = [
            p for p in paths
            if not os.path.basename(p).startswith('.')
        ]

    if basenames:
        paths = [p[len(folder) + 1:] for p in paths]

    return sorted(paths)


def mkdir(
        path: typing.Union[str, bytes],
        *,
        mode: int = 0o777
) -> str:
    """Create directory.

    Create a directory at the provided path
    and return the absolute path to the generated directory.
    If parent directories are missing,
    they are created as well.
    If the directory exists already,
    only its absolute path is returned.

    On some systems, mode is ignored.
    Where it is used, the current umask value is first masked out.
    If bits other than the last 9
    (i.e. the last 3 digits of the octal representation of the mode)
    are set,
    their meaning is platform-dependent.
    On some platforms, they are ignored
    and you should call :func:`os.chmod`
    explicitly to set them.

    Args:
        path: absolute or relative path of directory to create
        mode: set permissions of created folders

    Returns:
        absolute path to the created directory

    Examples:
        >>> p = mkdir('path1/path2/path3')
        >>> os.path.basename(p)
        'path3'

    """
    path = safe_path(path)
    if path:
        os.makedirs(path, mode=mode, exist_ok=True)
    return path


def md5(
        path: str,
        chunk_size: int = 8192,
) -> str:
    r"""Calculate MD5 checksum of file or folder.

    If ``path`` is a folder,
    the checksum is calculated
    from all files in the folder
    (including hidden files).
    The checksum also encodes
    the (relative) file names,
    so that renaming files
    results in a different checksum.
    However,
    the calculation is independent of
    file separator of the operation systems.
    Empty folders are ignored.

    Args:
        path: path to file or folder
        chunk_size: chunk size in number of bytes

    Returns:
        checksum

    Raises:
        FileNotFoundError: if ``path`` does not exist

    Examples:
        >>> path = touch('file.txt')
        >>> md5(path)
        'd41d8cd98f00b204e9800998ecf8427e'
        >>> md5('.')
        '3d8e577bddb17db339eae0b3d9bcf180'

    """
    path = safe_path(path)
    hasher = hashlib.md5()

    if not os.path.exists(path):
        raise FileNotFoundError(
            errno.ENOENT,
            os.strerror(errno.ENOENT),
            path,
        )

    if not os.path.isdir(path):

        with open(path, 'rb') as fp:
            for chunk in md5_read_chunk(fp, chunk_size):
                hasher.update(chunk)

    else:

        files = list_file_names(
            path,
            recursive=True,
            hidden=True,
            basenames=True,
        )

        for file in files:
            # encode file name that renaming of files
            # produces different checksum
            hasher.update(file.replace(os.path.sep, '/').encode())
            with open(safe_path(path, file), 'rb') as fp:
                for chunk in md5_read_chunk(fp, chunk_size):
                    hasher.update(chunk)

    return hasher.hexdigest()


def md5_read_chunk(
        fp: typing.IO,
        chunk_size: int = 8192,
):
    while True:
        data = fp.read(chunk_size)
        if not data:
            break
        yield data


def move_file(
        src_path,
        dst_path,
):
    """Move a file independent of operating system.

    As :func:`os.rename` works differently
    under Unix and Windows
    and :func:`shutil.move` can be slow,
    we use :func:`os.replace`
    to move the file.

    Args:
        scr_path: source file path
        dst_path: destination file path

    Examples:
        >>> path = mkdir('folder')
        >>> src_path = touch(os.path.join(path, 'file1'))
        >>> dst_path = os.path.join(path, 'file2')
        >>> move_file(src_path, dst_path)
        >>> list_file_names(path, basenames=True)
        ['file2']

    """
    os.replace(src_path, dst_path)


def replace_file_extension(
        path: typing.Union[str, bytes],
        new_extension: str,
        *,
        ext: str = None,
) -> str:
    """Replace file extension.

    If ``ext`` is ``None``
    it uses :func:`audeer.file_extension`
    to identify the current extension
    and replaces it with ``new_extension``.

    If ``ext`` is not ``None``
    but ``path`` ends on a different extension,
    the original path is returned.

    Args:
        path: path to file
        new_extension: new file extension
        ext: explicit extension to be removed

    Returns:
        path to file with a possibly new extension

    Examples:
        >>> replace_file_extension('file.txt', 'rst')
        'file.rst'
        >>> replace_file_extension('file', 'rst')
        'file.rst'
        >>> replace_file_extension('file.txt', '')
        'file'
        >>> replace_file_extension('file.tar.gz', 'zip', ext='tar.gz')
        'file.zip'
        >>> replace_file_extension('file.zip', 'rst', ext='txt')
        'file.zip'

    """
    if ext is None:
        ext = file_extension(path)

    # '.mp3' => 'mp3'
    if ext.startswith('.'):
        ext = ext[1:]
    if new_extension.startswith('.'):
        new_extension = new_extension[1:]

    if ext and not path.endswith(f'.{ext}'):
        return path

    if not path:
        return path

    if not ext and not new_extension:
        pass
    elif not ext:
        path = f'{path}.{new_extension}'
    elif not new_extension:
        path = path[:-len(ext) - 1]
    else:
        path = f'{path[:-len(ext)]}{new_extension}'

    return path


def rmdir(
        path: typing.Union[str, bytes],
):
    """Remove directory.

    Remove the directory
    and all its content
    if the directory exists.

    Args:
        path: absolute or relative path of directory to remove

    Raises:
        NotADirectoryError: if path is not a directory

    Examples:
        >>> _ = mkdir('path1/path2/path3')
        >>> list_dir_names('path1', basenames=True)
        ['path2']
        >>> rmdir('path1/path2')
        >>> list_dir_names('path1')
        []

    """
    path = safe_path(path)
    if os.path.exists(path):
        shutil.rmtree(path)


def touch(
        path: typing.Union[str, bytes]
) -> str:
    """Create an empty file.

    If the file exists already
    it's access and modification times
    are updated.

    Args:
        path: path to file

    Returns:
        expanded path to file

    Examples:
        >>> path = touch('file.txt')
        >>> os.path.basename(path)
        'file.txt'

    """
    path = safe_path(path)
    if os.path.exists(path):
        os.utime(path, None)
    else:
        open(path, 'a').close()
    return path



import operator
import pytest

import audeer


@pytest.mark.parametrize(
    'v1, v2, wanted',
    [
        ('1.5.1', '1.5.2b2', -1),
        ('161', '3.10a', 1),
        ('8.02', '8.02', 0),
        ('3.4j', '1996.07.12', -1),
        ('3.2.pl0', '3.1.1.6', 1),
        ('2g6', '11g', -1),
        ('0.960923', '2.2beta29', -1),
        ('1.13++', '5.5.kw', -1),
    ]
)
def test_loose_version(v1, v2, wanted):

    # This test is replicating the original tests from
    # https://github.com/python/cpython/blob/20a1c8ee4bcb1c421b7cca1f3f5d6ad7ce30a9c9/Lib/distutils/tests/test_version.py

    v1 = audeer.LooseVersion(v1)

    result = v1._cmp(audeer.LooseVersion(v2))
    assert result == wanted

    result = v1._cmp(v2)
    assert result == wanted

    result = v1._cmp(object())
    assert result == NotImplemented

    assert v1.__repr__() == f"LooseVersion ('{str(v1)}')"


@pytest.mark.parametrize(
    'v1, v2, operator, expected',
    [
        ('1.5.1', '1.5.2b2', operator.eq, False),
        ('1.5.1', '1.5.2b2', operator.lt, True),
        ('1.5.1', '1.5.2b2', operator.le, True),
        ('1.5.1', '1.5.2b2', operator.gt, False),
        ('1.5.1', '1.5.2b2', operator.ge, False),
        ('1.5.1', '1.5.1', operator.eq, True),
        ('1.5.1', '1.5.1', operator.le, True),
        ('1.5.1', '1.5.1', operator.ge, True),
        ('161', '3.10a', operator.gt, True),
        ('8.02', '8.02', operator.eq, True),
        ('3.4j', '1996.07.12', operator.lt, True),
        ('3.2.pl0', '3.1.1.6', operator.gt, True),
        ('2g6', '11g', operator.lt, True),
        ('0.960923', '2.2beta29', operator.lt, True),
        ('1.13++', '5.5.kw', operator.lt, True),
    ]
)
def test_loose_version_operator(v1, v2, operator, expected):
    v1 = audeer.LooseVersion(v1)
    v2 = audeer.LooseVersion(v2)
    result = operator(v1, v2)
    assert result == expected


@pytest.mark.parametrize(
    'v1, v2, wanted',
    [
        ('1.5.1', '1.5.2b2', -1),
        pytest.param(
            '161',
            '3.10a',
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        ('8.02', '8.02', 0),
        pytest.param(
            '3.4j',
            '1996.07.12',
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            '3.2.pl0',
            '3.1.1.6',
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            '2g6',
            '11g',
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        ('0.9', '2.2', -1),
        ('1.2.1', '1.2', 1),
        ('1.1', '1.2.2', -1),
        ('1.2', '1.1', 1),
        ('1.2.1', '1.2.2', -1),
        ('1.2.2', '1.2', 1),
        ('1.2', '1.2.2', -1),
        ('0.4.0', '0.4', 0),
        pytest.param(
            '1.13++',
            '5.5.kw',
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ]
)
def test_strict_version(v1, v2, wanted):

    # This test is replicating the original tests from
    # https://github.com/python/cpython/blob/20a1c8ee4bcb1c421b7cca1f3f5d6ad7ce30a9c9/Lib/distutils/tests/test_version.py

    v1 = audeer.StrictVersion(v1)

    result = v1._cmp(audeer.StrictVersion(v2))
    assert result == wanted

    result = v1._cmp(v2)
    assert result == wanted

    result = v1._cmp(object())
    assert result == NotImplemented

    assert v1.__repr__() == f"StrictVersion ('{str(v1)}')"


@pytest.mark.parametrize(
    'v1, v2, operator, expected',
    [
        ('1.5.1', '1.5.2b2', operator.eq, False),
        ('1.5.1', '1.5.2b2', operator.lt, True),
        ('1.5.1', '1.5.2b2', operator.le, True),
        ('1.5.1', '1.5.2b2', operator.gt, False),
        ('1.5.1', '1.5.2b2', operator.ge, False),
        ('1.5.1', '1.5.2b2', operator.ge, False),
        ('1.5.1a1', '1.5.1', operator.lt, True),
        ('1.5.1a1', '1.5.1a2', operator.lt, True),
        ('1.5.1a2', '1.5.1a1', operator.gt, True),
        ('1.5.1a1', '1.5.1a1', operator.eq, True),
        ('1.5.1', '1.5.1b2', operator.gt, True),
        ('1.5.1', '1.5.1', operator.le, True),
        ('1.5.1', '1.5.1', operator.ge, True),
        ('8.02', '8.02', operator.eq, True),
        ('0.9', '2.2', operator.lt, True),
        ('1.2.1', '1.2', operator.gt, True),
        ('1.1', '1.2.2', operator.lt, True),
        ('1.2', '1.1', operator.gt, True),
        ('1.2.1', '1.2.2', operator.lt, True),
        ('1.2.2', '1.2', operator.gt, True),
        ('1.2', '1.2.2', operator.lt, True),
        ('0.4.0', '0.4', operator.eq, True),
    ]
)
def test_strict_version_operator(v1, v2, operator, expected):
    v1 = audeer.StrictVersion(v1)
    v2 = audeer.StrictVersion(v2)
    result = operator(v1, v2)
    assert result == expected
